"""."""

from association_studio.app.back import AssociationStudioBackend
from association_studio.app.show_coo_mode import ShowCooMode


def test_next_show_coo_mode(be: AssociationStudioBackend) -> None:
    be.next_show_coo_mode()
    assert be.show_coo_mode == ShowCooMode.TRANSLATION



"""."""

from typing import Sequence


class MetricMode:
    def __init__(self, metric_names: Sequence[str]) -> None:
        self.choices = [str(name) for name in metric_names]
        self.metric = self.choices[0]

    def next(self) -> None:
        i = (self.choices.index(self.metric) + 1) % len(self.choices)
        self.metric = self.choices[i]



"""."""

import matplotlib.pyplot as plt
import pytest

from matplotlib.backend_bases import MouseEvent, PickEvent

from association_studio.app.back import AssociationStudioBackend


def test_translate(ax: plt.Axes, be: AssociationStudioBackend) -> None:
    canvas = ax.figure.canvas
    mouse_event_pick = MouseEvent('mouse-pick', canvas, 100, 200)
    pick_event = PickEvent('pick', ax.figure.canvas, mouse_event_pick, be.probe_poly)
    mouse_event_move = MouseEvent('mouse-move', canvas, 110, 210)
    be.change_trk_pose(pick_event, mouse_event_move)
    assert be.box_probe.vec_z[:3] == pytest.approx([0.32016129, 0.72705628, 0.0])
    assert be.box_probe.vec_z[3:7] == pytest.approx(
        [0.9284766908852592, 0.0, 0.0, -0.3713906763541037]
    )
    assert be.box_probe.vec_z[7:] == pytest.approx([4.0, 2.0, 1.0])


def test_rotate(ax: plt.Axes, be: AssociationStudioBackend) -> None:
    canvas = ax.figure.canvas
    mouse_event_pick = MouseEvent('mouse-pick', canvas, 100, 200)
    pick_event = PickEvent('pick', ax.figure.canvas, mouse_event_pick, be.probe_circle)
    mouse_event_move = MouseEvent('mouse-move', canvas, 110, 210)
    be.change_trk_pose(pick_event, mouse_event_move)
    assert be.box_probe.vec_z[:3] == pytest.approx([0.3, 0.7, 0.0])
    assert be.box_probe.vec_z[3:7] == pytest.approx([0.92726528078645, 0.0, 0.0, -0.37440499335348])
    assert be.box_probe.vec_z[7:] == pytest.approx([4.0, 2.0, 1.0])



"""."""

import pytest

from pytest_mock import MockerFixture

from association_studio.app.back import AssociationStudioBackend


def test_set_size_probe(be: AssociationStudioBackend, mocker: MockerFixture) -> None:
    mock_upd = mocker.patch.object(be, 'upd_scores_and_probe')
    be.set_size_probe(10.0, 5.0)
    assert be.box_probe.vec_z[7:] == pytest.approx((10.0, 5.0, 1.0))
    assert be.box_ref.vec_z[7:] == pytest.approx((4.0, 2.0, 1.0))
    mock_upd.assert_called_once()



import matplotlib.pyplot as plt
import pytest

from matplotlib.backend_bases import KeyEvent, MouseEvent, PickEvent
from pytest_mock import MockerFixture

from association_studio.app.back import AssociationStudioBackend
from association_studio.app.front import AssociationFrontend, PltHolder


@pytest.fixture
def front(mocker: MockerFixture) -> AssociationFrontend:
    mocker.patch.object(PltHolder, 'ion')
    return AssociationFrontend(AssociationStudioBackend())


@pytest.fixture
def pick(ax: plt.Axes, front: AssociationFrontend) -> PickEvent:
    mouse_pick = MouseEvent('mouse pick', ax.figure.canvas, 100, 200)
    return PickEvent('my pick', ax.figure.canvas, mouse_pick, front.back.probe_poly)


@pytest.fixture
def key_event(ax: plt.Axes) -> KeyEvent:
    return KeyEvent('key event', ax.figure.canvas, 'c')



"""."""

import cv2
import numpy as np
import pytest

from association_studio.exp_metrics.iou_like_dia_mah import IouLikeDiaMah


def test_compute_metric(driver: IouLikeDiaMah, kf: cv2.KalmanFilter) -> None:
    det_1z = np.array((0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 4.0, 2.0, 1.0)).reshape(1, 10)
    with pytest.raises(RuntimeWarning):
        driver.compute_metric(det_1z, [kf])



"""."""

import pytest

from pytest_mock import MockerFixture

from association_studio.app.back import AssociationStudioBackend


def test_rotate_probe_30(be: AssociationStudioBackend, mocker: MockerFixture) -> None:
    mock_upd = mocker.patch.object(be, 'upd_scores_and_probe')
    be.rotate_probe_30()
    ref = 0.9929625950474326, 0.0, 0.0, -0.11842839539852248
    assert be.box_probe.vec_z[3:7] == pytest.approx(ref)
    mock_upd.assert_called_once()



"""."""

from matplotlib.backend_bases import KeyEvent

from association_studio.app.front import AssociationFrontend
from association_studio.app.show_coo_mode import ShowCooMode


def test_coo_mode(front: AssociationFrontend, key_event: KeyEvent) -> None:
    front.on_key(key_event)
    assert front.back.show_coo_mode == ShowCooMode.TRANSLATION
    assert front.text.get_text() == 'xy 0.300 0.700'



"""."""

from typing import Sequence


def get_labels(names: Sequence[str]) -> list[str]:
    """."""
    labels = []
    for name in names:
        parts = name.split('-')
        one_or_two = [parts[0], '-'.join(parts[1:])] if len(parts) > 1 else parts
        labels.append('\n'.join(one_or_two))
    return labels



"""."""

import pytest

from pytest_mock import MockerFixture

from association_studio.app.back import AssociationStudioBackend


def test_move_probe_to_origin(be: AssociationStudioBackend, mocker: MockerFixture) -> None:
    mock_upd = mocker.patch.object(be, 'upd_probe_patches')
    be.move_probe_to_origin()
    assert be.score_driver.scores == pytest.approx(1.0)
    mock_upd.assert_called_once()



import matplotlib.pyplot as plt


def get_fig_axes() -> tuple[plt.Figure, plt.Axes]:
    fig, ax = plt.subplots(figsize=(5, 2.5))
    fig.subplots_adjust(left=0.3)
    fig.canvas.manager.set_window_title('Association Studio: metrics')
    return fig, ax



"""."""

import pytest

from matplotlib.backend_bases import KeyEvent
from pytest_mock import MockerFixture

from association_studio.app.front import AssociationFrontend


def test_origin(front: AssociationFrontend, key_event: KeyEvent, mocker: MockerFixture) -> None:
    key_event.key = 'O'
    mock_update = mocker.patch.object(front.back, 'update_horizontal_bar_container')
    front.on_key(key_event)
    mock_update.assert_called_once_with(front.bar)
    assert front.back.score_driver.scores == pytest.approx(1.0)
    assert front.ax_boxes.get_title() == 'giou-axes-aligned 1.0000'



"""."""

import pytest

from association_studio.app.metric_mode import MetricMode


@pytest.fixture
def mm() -> MetricMode:
    return MetricMode(['1', '2', '3', '4'])


def test_metric_mode(mm: MetricMode) -> None:
    assert len(mm.choices) > 1
    assert mm.choices[0] == '1'
    assert mm.metric == mm.choices[0]


def test_next(mm: MetricMode) -> None:
    mm.next()
    assert mm.metric == mm.choices[1]
    mm.next()
    assert mm.metric == mm.choices[2]
    mm.next()
    assert mm.metric == mm.choices[3]
    mm.next()
    assert mm.metric == mm.choices[0]



"""."""

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.backend_bases import Event, KeyEvent, MouseEvent, PickEvent

from association_studio.mpl_vis.horizontal_bar import get_fig_axes
from association_studio.mpl_vis.prepare_axes import prepare_axes

from .back import AssociationStudioBackend


class PltHolder:
    ion = plt.ion
    show = plt.show


class AssociationFrontend:
    def __init__(self, back: AssociationStudioBackend) -> None:
        self.back = back
        PltHolder.ion()  # replot on any sneeze
        self.fig_boxes, self.ax_boxes = plt.subplots()
        self.fig_boxes.canvas.manager.set_window_title('Association Studio: boxes')
        self.ax_boxes.set_title(self.back.get_title())
        prepare_axes(self.ax_boxes, np.array([(-15.0, -15.0), (5.0, 5.0)]))
        self.ax_boxes.add_artist(self.back.ref_poly)
        self.ax_boxes.add_artist(self.back.ref_circle)
        self.ax_boxes.add_artist(self.back.probe_poly)
        self.ax_boxes.add_artist(self.back.probe_circle)
        bbox_dct = dict(boxstyle='round,pad=0.5', fc='wheat', ec='orange')
        self.text = self.ax_boxes.text(-14.0, -14.0, back.get_text(), size=10, bbox=bbox_dct)
        self.fig_boxes.canvas.mpl_connect('key_press_event', self.on_key)
        self.fig_boxes.canvas.mpl_connect('button_release_event', self.on_release)
        self.fig_boxes.canvas.mpl_connect('motion_notify_event', self.on_move)
        self.fig_boxes.canvas.mpl_connect('pick_event', self.on_pick)
        self.pick_events: list[PickEvent] = []
        self.fig_metrics, self.ax_metrics = get_fig_axes()
        self.bar = back.get_horizontal_bar_container(self.ax_metrics)

    def on_key(self, event: Event | KeyEvent) -> None:
        match event.key:
            case 'c':
                self.back.next_show_coo_mode()
            case 'm':
                self.back.next_metric()
            case 'O':
                self.back.move_probe_to_origin()
            case 'R':
                self.back.rotate_probe_30()
            case '1':
                self.back.set_unit_size_ref()
            case '2':
                self.back.set_size_probe(2.0, 4.0)

        self.back.update_horizontal_bar_container(self.bar)
        self.ax_boxes.set_title(self.back.get_title())
        self.text.set_text(self.back.get_text())

    def on_pick(self, event: Event | PickEvent) -> None:
        self.pick_events.append(event)
        self.back.start_pose_change()

    def on_release(self, _event: Event | MouseEvent) -> None:
        self.pick_events.clear()

    def on_move(self, event: Event | MouseEvent) -> None:
        if self.pick_events and event.xdata is not None:
            self.back.change_trk_pose(self.pick_events[-1], event)
            self.text.set_text(self.back.get_text())
            self.ax_boxes.set_title(self.back.get_title())
            self.back.update_horizontal_bar_container(self.bar)

    @staticmethod
    def show() -> None:
        PltHolder.show(block=True)



"""."""

import cv2
import numpy as np

from kinematic_tracker.association.metric_driver_base import MetricDriverBase

from association_studio.common import FMT


class IouLikeDiaMah(MetricDriverBase):
    NAME = 'iou-like-dia-mah'

    def __init__(self) -> None:
        super().__init__(1, 1, 10)
        self.pos = 0.308
        self.size = 1.0
        self.qua = 2.0
        self._precision_dia = np.zeros(10)
        self._precision_dia[3:7] = self.qua
        self._square_size = np.zeros(3)
        self._square_diff_z = np.zeros(10)

    def compute_metric(self, det_1z: FMT, filters: list[cv2.KalmanFilter]) -> FMT:
        if not np.allclose(det_1z[0, 4:6], 0.0):
            raise RuntimeWarning('rx, ry are assumed to be zeros')

        metric_rt = self.get_rect_chunk(1, 1)
        ref_z = det_1z[0]
        probe_z = filters[0].statePre[:, 0]
        self._square_size[:] = np.square((ref_z[7:] + probe_z[7:]) / 2.0)
        self._precision_dia[:3] = self.pos / self._square_size
        self._precision_dia[7:] = self.size / self._square_size
        self._square_diff_z[:] = np.square(probe_z - ref_z)
        square_distance = float(np.sum(self._precision_dia * self._square_diff_z))
        metric_rt[0, 0] = np.exp(-square_distance / 2)
        return metric_rt



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw

from association_studio.common import FMT


@pytest.fixture
def ref_drv() -> MetricGIoUYaw:
    indices = 0, 1, 2, 3, 6, 7, 8, 9
    return MetricGIoUYaw(1, 1, 10, indices)


@pytest.fixture
def kf() -> cv2.KalmanFilter:
    kf = cv2.KalmanFilter(10, 10, 0, cv2.CV_64F)
    kf.statePre = np.array((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 4.0, 2.0, 1.0)).reshape(10, 1)
    return kf


def get_kf(pre_z: FMT) -> cv2.KalmanFilter:
    kf = cv2.KalmanFilter(10, 10, 0, cv2.CV_64F)
    kf.statePre = pre_z.reshape(10, 1)
    kf.measurementMatrix = np.eye(10)
    return kf


@pytest.fixture
def ref_1z() -> FMT:
    return np.array((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0)).reshape(1, 10)



"""."""

from enum import Enum

from association_studio.core.mid_box import Box


class ShowCooMode(Enum):
    TRANSLATION_AND_ROTATION = 0
    TRANSLATION = 1
    ROTATION = 2
    NOTHING = 3

    def next(self) -> int:
        p1 = self.value + 1
        return 0 if p1 > 3 else p1

    def format(self, box: Box) -> str:
        if self == self.TRANSLATION_AND_ROTATION:
            x, y = box.vec_z[:2]
            w, rz = box.vec_z[3:7:3]
            msg = f'xy {x:5.3f} {y:5.3f} w rz {w:5.3f} {rz:5.3f}'
        elif self == self.TRANSLATION:
            x, y = box.vec_z[:2]
            msg = f'xy {x:5.3f} {y:5.3f}'
        elif self == self.ROTATION:
            w, rz = box.vec_z[3:7:3]
            msg = f'w rz {w:5.3f} {rz:5.3f}'
        else:
            msg = ''
        return msg



"""."""

import numpy as np
import pytest

from association_studio.app.back import AssociationStudioBackend
from association_studio.core.mid_box import Box


def test_upd_probe_patches(be: AssociationStudioBackend, box: Box) -> None:
    be.box_probe = box
    be.upd_probe_patches()
    # fmt: off
    ref = [
        [2.6644295270646188, 4.7348993307680365], [-2.0335570481537326, 3.0234899294066446],
        [-0.6644295270646188, -0.7348993307680365], [4.033557048153733, 0.9765100705933556],
        [2.6644295270646188, 4.7348993307680365]
    ]
    # fmt: on
    assert be.probe_poly.get_xy() == pytest.approx(np.array(ref))
    assert be.probe_circle.get_center() == pytest.approx([1.684563760544557, 0.12080536991265955])



import numpy as np
import scipy.spatial.transform

from association_studio.common import FMT, FVT


MID_CORNERS_CN = ((0.5, 0.5, -0.5, -0.5), (-0.5, 0.5, 0.5, -0.5), (0.0, 0.0, 0.0, 0.0))


class Box:
    """Bounding box."""

    def __init__(self, vec_z: FVT) -> None:
        self._rotation = scipy.spatial.transform.Rotation.from_quat(vec_z[3:7], scalar_first=True)
        self.vec_z = vec_z
        self.unit_corners = np.array(MID_CORNERS_CN)
        self.scaled_corners = np.zeros_like(self.unit_corners)
        self.corners_buf = np.zeros_like(self.unit_corners)
        self._wlh31 = np.zeros((3, 1))

    def set_size(self, width: float, length: float, height: float) -> None:
        self.vec_z[7:] = width, length, height

    def translate(self, x: FVT) -> None:
        self.vec_z[:3] += x

    def rotate(self, wxyz: FVT) -> None:
        self._rotation *= scipy.spatial.transform.Rotation.from_quat(wxyz, scalar_first=True)
        self.vec_z[3:7] = self._rotation.as_quat(scalar_first=True, canonical=True)

    def mid_corners(self) -> FMT:
        self._wlh31[:, 0] = self.vec_z[7:]
        self.scaled_corners[:] = self.unit_corners
        self.scaled_corners *= self._wlh31
        np.dot(self._rotation.as_matrix(), self.scaled_corners, out=self.corners_buf)
        self.corners_buf += self.vec_z[:3].reshape(3, 1)
        return self.corners_buf



"""."""

import pytest

from matplotlib.backend_bases import KeyEvent
from pytest_mock import MockerFixture

from association_studio.app.front import AssociationFrontend


def test_set_unit(front: AssociationFrontend, key_event: KeyEvent, mocker: MockerFixture) -> None:
    key_event.key = '1'
    mock_update = mocker.patch.object(front.back, 'update_horizontal_bar_container')
    front.on_key(key_event)
    mock_update.assert_called_once_with(front.bar)
    ref = 0.5146895787, 0.51274987966, 0.764849523777, 0.90554415261, 0.325945784, 0.46321533
    assert front.back.score_driver.scores == pytest.approx(ref)
    assert front.ax_boxes.get_title() == 'giou-axes-aligned 0.5147'



"""."""

import numpy as np

from kinematic_tracker.association.association_metric import AssociationMetric

from association_studio.common import FVT, OVT

from .tracker_factory import get_low_order_nu_scenes_tracker
from .valid_metrics import ValidMetrics


class ScoreDriver:
    def __init__(self) -> None:
        self.valid_metrics = ValidMetrics()
        self.probe_1z = np.zeros((1, 10))
        tracker = get_low_order_nu_scenes_tracker()
        self.ref_1z = np.array([(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 4.0, 2.0, 1.0)])
        tracker.advance(0, self.ref_1z)
        self._ref_kfs = [tracker.tracks_c[0][0].kkf.kalman_filter]
        self.metric_names = np.array(self.valid_metrics.get_valid_metrics(), dtype=str)
        num_metrics = len(self.metric_names)
        self.metric_drivers: OVT = np.zeros(num_metrics, object)
        self.scores = np.zeros(num_metrics)
        self._init_metric_drivers()
        self.metric_mask = np.zeros(num_metrics, bool)
        self.metric_index = -1
        self.score = 0.0
        self.set_association_metric(AssociationMetric.MAHALANOBIS_SIZE_MODULATED.value)

    def _init_metric_drivers(self) -> None:
        for metric_num, metric_name in enumerate(self.metric_names):
            self.metric_drivers[metric_num] = self.valid_metrics.get_driver(metric_name)

    def set_association_metric(self, metric: str) -> None:
        self.metric_mask[:] = metric == self.metric_names
        self.metric_index = int(np.argmax(self.metric_mask))

    def set_ref_z(self, ref_z: FVT) -> None:
        self._ref_kfs[0].statePre[:, 0] = ref_z

    def compute_score(self, vec_z: FVT) -> None:
        self.compute_scores(vec_z)
        assert self.metric_index >= 0
        self.score = self.scores[self.metric_index]

    def compute_scores(self, vec_z: FVT) -> None:
        self.probe_1z[0] = vec_z
        for idx, driver in enumerate(self.metric_drivers):
            self.scores[idx] = driver.compute_metric(self.probe_1z, self._ref_kfs)[0, 0]



"""."""

import numpy as np

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw

from association_studio.exp_metrics.iou_like_dia_mah import IouLikeDiaMah

from ..conftest import get_kf


def test_rotation(driver: IouLikeDiaMah, ref_drv: MetricGIoUYaw) -> None:
    """."""
    ref_1z = np.array((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 4.0, 2.0, 1.0)).reshape(1, 10)
    probe90_z = np.array((0.0, 0.0, 0.0, 0.707106781, 0.0, 0.0, 0.707106781, 4.0, 2.0, 1.0))
    kf90 = get_kf(probe90_z)
    ref_90 = ref_drv.compute_metric(ref_1z, [kf90])[0, 0]
    probe_90 = driver.compute_metric(ref_1z, [kf90])[0, 0]
    assert abs(probe_90 - ref_90) < 0.04



"""."""

from association_studio.score_driver.score_driver import ScoreDriver


def test_set_association_metric(driver: ScoreDriver) -> None:
    assert driver.metric_index == 3
    driver.set_association_metric('mahalanobis')
    assert driver.metric_index == 2



"""."""

import matplotlib.pyplot as plt
import pytest

from association_studio.mpl_vis.horizontal_bar import get_fig_axes


def test_get_fig_axes():
    fig, ax = get_fig_axes()
    assert isinstance(fig, plt.Figure)
    assert isinstance(ax, plt.Axes)
    assert fig.get_figwidth() == pytest.approx(5.0)
    assert fig.get_figheight() == pytest.approx(2.5)
    assert fig.get_dpi() == pytest.approx(100.0)
    assert fig.canvas.manager.get_window_title() == 'Association Studio: metrics'



"""."""

from kinematic_tracker.association.association import MDT
from kinematic_tracker.association.association_metric import AssociationMetric

from association_studio.exp_metrics.iou_like_dia_mah import IouLikeDiaMah
from association_studio.exp_metrics.iou_like_dia_slater import IouLikeDiaSlater

from .tracker_factory import get_low_order_nu_scenes_tracker


class ValidMetrics:
    """."""

    def __init__(self) -> None:
        self.tracker = get_low_order_nu_scenes_tracker()
        self.kt_metrics = [mt.value for mt in AssociationMetric]

    def get_valid_metrics(self) -> list[str]:
        return self.kt_metrics + [IouLikeDiaMah.NAME, IouLikeDiaSlater.NAME]

    def get_driver(self, name: str) -> MDT | IouLikeDiaMah | IouLikeDiaSlater:
        match name:
            case m if m in self.kt_metrics:
                self.tracker.set_association_metric(name)
                driver = self.tracker.metric_driver
            case IouLikeDiaMah.NAME:
                driver = IouLikeDiaMah()
            case IouLikeDiaSlater.NAME:
                driver = IouLikeDiaSlater()
            case _:
                raise ValueError(f'Unknown metric {name}')
        return driver



"""."""

from matplotlib.backend_bases import MouseEvent, PickEvent
from pytest_mock import MockerFixture

from association_studio.app.front import AssociationFrontend


def test_no_picks(front: AssociationFrontend, mocker: MockerFixture) -> None:
    mouse_event = MouseEvent('mouse move', front.ax_boxes.figure.canvas, 100, 200)
    mock_chg = mocker.patch.object(front.back, 'change_trk_pose')
    front.on_move(mouse_event)
    mock_chg.assert_not_called()


def test_1_pick(front: AssociationFrontend, mocker: MockerFixture, pick: PickEvent) -> None:
    mouse_event = MouseEvent('mouse move', front.ax_boxes.figure.canvas, 100, 200)
    mock_chg = mocker.patch.object(front.back, 'change_trk_pose')
    mock_upd_bar = mocker.patch.object(front.back, 'update_horizontal_bar_container')
    front.pick_events.append(pick)
    front.on_move(mouse_event)
    mock_chg.assert_called_once_with(pick, mouse_event)
    mock_upd_bar.assert_called_once_with(front.bar)



"""."""

import pytest

from matplotlib.backend_bases import KeyEvent
from pytest_mock import MockerFixture

from association_studio.app.front import AssociationFrontend


def test_rotate(front: AssociationFrontend, key_event: KeyEvent, mocker: MockerFixture) -> None:
    key_event.key = 'R'
    mock_update = mocker.patch.object(front.back, 'update_horizontal_bar_container')
    front.on_key(key_event)
    mock_update.assert_called_once_with(front.bar)
    ref = 0.69683618402, 0.7073877687, 0.9852578748, 0.993097065, 0.96675897987, 0.8307008
    assert front.back.score_driver.scores == pytest.approx(ref)
    assert front.ax_boxes.get_title() == 'giou-axes-aligned 0.6968'



"""."""

import pytest

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw

from association_studio.exp_metrics.iou_like_dia_mah import IouLikeDiaMah
from association_studio.exp_metrics.iou_like_dia_slater import IouLikeDiaSlater
from association_studio.score_driver.valid_metrics import ValidMetrics


@pytest.fixture
def vms() -> ValidMetrics:
    return ValidMetrics()


def test_init(vms: ValidMetrics) -> None:
    assert isinstance(vms.tracker, NdKkfTracker)


def test_get_valid_metrics(vms: ValidMetrics) -> None:
    # fmt: off
    ref = [
        'giou-axes-aligned', 'giou-yaw',
        'mahalanobis', 'mahalanobis-size-modulated',
        'iou-like-dia-mah', 'iou-like-dia-slater'
    ]
    # fmt: on
    assert vms.get_valid_metrics() == ref


def test_driver(vms: ValidMetrics) -> None:
    assert isinstance(vms.get_driver('giou-yaw'), MetricGIoUYaw)
    assert isinstance(vms.get_driver('iou-like-dia-mah'), IouLikeDiaMah)
    assert isinstance(vms.get_driver('iou-like-dia-slater'), IouLikeDiaSlater)
    with pytest.raises(ValueError):
        vms.get_driver('bogus-metric')



"""."""

import matplotlib.pyplot as plt
import pytest

from association_studio.app.back import AssociationStudioBackend


def test_it(be: AssociationStudioBackend, ax: plt.Axes) -> None:
    bar = ax.barh(range(4), range(4))
    be.update_horizontal_bar_container(bar)
    scores = [float(b.get_width()) for b in bar]
    ref = 0.6968361840236645, 0.6389961679998885, 0.9820862284712972, 0.9899001837330043
    assert scores == pytest.approx(ref)



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker

from association_studio.score_driver.tracker_factory import get_low_order_nu_scenes_tracker


def test_it() -> None:
    tracker = get_low_order_nu_scenes_tracker()
    assert isinstance(tracker, NdKkfTracker)
    assert tracker.shape.orders_x.tolist() == [1, 1, 1]
    assert tracker.cov_zz == pytest.approx(np.eye(10))



"""."""

from kinematic_tracker.association.association_metric import AssociationMetric
from matplotlib.backend_bases import KeyEvent

from association_studio.app.front import AssociationFrontend


def test_metric(front: AssociationFrontend, key_event: KeyEvent) -> None:
    key_event.key = 'm'
    front.on_key(key_event)
    assert front.back.metric_mode.metric == AssociationMetric.GIOU_YAW.value
    assert front.ax_boxes.get_title() == 'giou-yaw 0.6390'



"""."""

import pytest

from pytest_mock import MockerFixture

from association_studio.app.back import AssociationStudioBackend
from association_studio.core.mid_box import Box


def test_upd_scores_and_probe(
    be: AssociationStudioBackend, mocker: MockerFixture, box: Box
) -> None:
    be.box_probe = box
    mock_upd_probe_patches = mocker.patch.object(be, 'upd_probe_patches')
    mock_upd_ref_patches = mocker.patch.object(be, 'upd_ref_patches')
    be.upd_scores_and_probe()
    mock_upd_probe_patches.assert_called_once()
    mock_upd_ref_patches.assert_called_once()
    ref = 0.3606945738, 0.3788795, 0.294838691, 0.7805876354, 0.0894671298, 0.31149245
    assert be.score_driver.scores == pytest.approx(ref)



"""."""

import copy

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.backend_bases import MouseEvent, PickEvent
from matplotlib.container import BarContainer
from matplotlib.patches import Circle, Polygon
from scipy.spatial.transform import Rotation

from association_studio.core.mid_box import Box
from association_studio.score_driver.metric_labels import get_labels
from association_studio.score_driver.score_driver import ScoreDriver

from .metric_mode import MetricMode
from .show_coo_mode import ShowCooMode


class AssociationStudioBackend:
    def __init__(self) -> None:
        self.score_driver = ScoreDriver()
        self.show_coo_mode = ShowCooMode.TRANSLATION_AND_ROTATION
        self.metric_mode = MetricMode(self.score_driver.metric_names)
        self.score_driver.set_association_metric(self.metric_mode.metric)
        self.box_ref = Box(self.score_driver.ref_1z[0])
        t_box = 0.3, 0.7, 0.0, 0.928476691, 0.0, 0.0, -0.371390676, 4.0, 2.0, 1.0
        self.box_probe = Box(np.array(t_box))
        self.box_probe_start_pose_change = copy.deepcopy(self.box_probe)
        bcc = self.box_ref.mid_corners()[:2, :].T
        self.ref_poly = Polygon(bcc, edgecolor='k', facecolor='yellow', closed=True)
        self.ref_circle = Circle(bcc[2:].mean(axis=0), 0.35, edgecolor='k', facecolor='yellow')

        bcc = self.box_probe.mid_corners()[:2, :].T
        self.probe_poly = Polygon(bcc, edgecolor='b', facecolor='none', closed=True, picker=True)
        trk_xy = bcc[2:].mean(axis=0)
        self.probe_circle = Circle(trk_xy, 0.35, edgecolor='b', facecolor='w', picker=True)
        self.score_driver.compute_score(self.box_probe.vec_z)
        self._rotate_30_quat = np.array((0.9659258262890683, 0.0, 0.0, 0.25881904510252074))

    def next_show_coo_mode(self) -> None:
        self.show_coo_mode = ShowCooMode(self.show_coo_mode.next())

    def next_metric(self) -> None:
        self.metric_mode.next()
        self.score_driver.set_association_metric(self.metric_mode.metric)
        self.score_driver.compute_score(self.box_probe.vec_z)

    def move_probe_to_origin(self) -> None:
        self.box_probe = copy.deepcopy(self.box_ref)
        self.upd_scores_and_probe()

    def rotate_probe_30(self) -> None:
        self.box_probe.rotate(self._rotate_30_quat)
        self.upd_scores_and_probe()

    def set_unit_size_ref(self) -> None:
        self.box_ref.set_size(1.0, 1.0, 1.0)
        self.upd_scores_and_probe()

    def set_size_probe(self, width: float, length: float) -> None:
        self.box_probe.set_size(width, length, 1.0)
        self.upd_scores_and_probe()

    def start_pose_change(self) -> None:
        self.box_probe_start_pose_change = copy.deepcopy(self.box_probe)

    def change_trk_pose(self, pick: PickEvent, event: MouseEvent) -> None:
        me = pick.mouseevent

        if pick.artist == self.probe_poly:
            translation = np.array((event.xdata - me.xdata, event.ydata - me.ydata, 0.0))
            self.box_probe = copy.deepcopy(self.box_probe_start_pose_change)
            self.box_probe.translate(translation)
        elif pick.artist == self.probe_circle:
            self.box_probe = copy.deepcopy(self.box_probe_start_pose_change)
            c = self.box_probe.vec_z[:3]
            a = np.array((event.xdata - c[0], event.ydata - c[1], 0.0))
            b = np.array((me.xdata - c[0], me.ydata - c[1], 0.0))
            q_ab = Rotation.align_vectors(a, b)[0].as_quat(scalar_first=True)
            self.box_probe.rotate(q_ab)

        self.upd_scores_and_probe()

    def upd_scores_and_probe(self) -> None:
        self.score_driver.set_ref_z(self.box_ref.vec_z)
        self.score_driver.compute_score(self.box_probe.vec_z)
        self.upd_probe_patches()
        self.upd_ref_patches()

    def upd_probe_patches(self) -> None:
        bcc = self.box_probe.mid_corners()[:2, :].T
        self.probe_poly.set_xy(bcc)
        self.probe_circle.set_center(bcc[2:].mean(axis=0))

    def upd_ref_patches(self) -> None:
        bcc = self.box_ref.mid_corners()[:2, :].T
        self.ref_poly.set_xy(bcc)
        self.ref_circle.set_center(bcc[2:].mean(axis=0))

    def get_title(self) -> str:
        return f'{self.metric_mode.metric} {self.score_driver.score:5.4f}'

    def get_text(self) -> str:
        return self.show_coo_mode.format(self.box_probe)

    def get_horizontal_bar_container(self, ax_metrics: plt.Axes) -> BarContainer:
        scores = self.score_driver.scores
        names = get_labels(self.score_driver.metric_names)
        return ax_metrics.barh(range(scores.shape[0]), scores, tick_label=names)

    def update_horizontal_bar_container(self, bar: BarContainer) -> None:
        for bar, new_h in zip(bar, self.score_driver.scores):
            bar.set_width(new_h)



"""."""

import os

from pathlib import Path
from typing import Generator

import matplotlib.pyplot as plt
import pytest


@pytest.fixture
def files_dir() -> Path:
    """."""
    return Path(__file__).parent / 'files'


@pytest.fixture(autouse=True)
def cd_tmp_path(tmp_path: Path) -> Generator[Path, None, None]:
    """."""
    orig_path = Path.cwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(orig_path)


@pytest.fixture
def ax() -> plt.Axes:
    ax = plt.gca()
    ax.clear()
    yield ax
    plt.close()



"""."""

import pytest

from association_studio.core.mid_box import Box


def test_set_size(box: Box) -> None:
    box.set_size(1.23, 3.45, 6.78)
    assert box.vec_z[7:] == pytest.approx((1.23, 3.45, 6.78))



"""."""

from association_studio.app.front import AssociationFrontend


def test_on_release(front: AssociationFrontend) -> None:
    """."""
    front.pick_events.append('anything here')
    front.on_release('Anything')
    assert front.pick_events == []



"""."""

from matplotlib.backend_bases import PickEvent
from pytest_mock import MockerFixture

from association_studio.app.front import AssociationFrontend


def test_on_pick(front: AssociationFrontend, mocker: MockerFixture, pick: PickEvent) -> None:
    mock_start = mocker.patch.object(front.back, 'start_pose_change')
    front.on_pick(pick)
    assert front.pick_events == [pick]
    mock_start.assert_called_once()



"""."""

import numpy as np

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw

from association_studio.common import FMT
from association_studio.exp_metrics.iou_like_dia_mah import IouLikeDiaMah

from ..conftest import get_kf


def test_scale(driver: IouLikeDiaMah, ref_drv: MetricGIoUYaw, ref_1z: FMT) -> None:
    """."""
    probe_z = np.array((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 2.0, 2.0, 1.0))
    kf = get_kf(probe_z)
    ref = ref_drv.compute_metric(ref_1z, [kf])[0, 0]
    probe = driver.compute_metric(ref_1z, [kf])[0, 0]
    assert abs(probe - ref) < 0.017



"""."""

import pytest

from association_studio.common import FVT
from association_studio.score_driver.score_driver import ScoreDriver


def test_compute_scores(vec_z: FVT, driver: ScoreDriver) -> None:
    driver.compute_scores(vec_z)
    ref = 0.139487179487, 0.1902341996, 0.252306998, 0.728156653, 0.181316789, 0.31068506
    assert driver.scores == pytest.approx(ref)



"""."""

from association_studio.app.front import AssociationFrontend


def test_init(front: AssociationFrontend) -> None:
    assert front.pick_events == []
    assert front.fig_boxes.canvas.manager.get_window_title() == 'Association Studio: boxes'
    assert front.ax_boxes.get_title() == 'giou-axes-aligned 0.6968'
    assert len(front.ax_metrics.patches) == 6
